"""FABulous FPGA framework package.

This file marks the FABulous directory as a Python package so Sphinx and tools can
import modules (e.g., FABulous.fabric_definition) during documentation builds.
"""
